<?php

declare(strict_types=1);

/*
 * Copyright (c) 2017-2022 François Kooman <fkooman@tuxed.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

namespace fkooman\OAuth\Server;

use DateTimeImmutable;
use fkooman\OAuth\Server\Exception\InvalidGrantException;

class RefreshToken
{
    public const TOKEN_TYPE = 'refresh_token';

    private string $tokenId;
    private string $authKey;
    private string $userId;
    private string $clientId;
    private Scope $scope;
    private DateTimeImmutable $expiresAt;

    public function __construct(string $tokenId, string $authKey, string $userId, string $clientId, Scope $scope, DateTimeImmutable $expiresAt)
    {
        $this->tokenId = $tokenId;      // jti
        $this->authKey = $authKey;      // ???
        $this->userId = $userId;        // sub
        $this->clientId = $clientId;    // client_id
        $this->scope = $scope;          // scope
        $this->expiresAt = $expiresAt;  // exp
    }

    public static function fromJson(string $jsonString): self
    {
        $jsonData = Json::decode($jsonString);
        if (!isset($jsonData['type']) || !\is_string($jsonData['type']) || self::TOKEN_TYPE !== $jsonData['type']) {
            throw new InvalidGrantException('missing, malformed or invalid "type"');
        }
        if (!isset($jsonData['token_id']) || !\is_string($jsonData['token_id'])) {
            throw new InvalidGrantException('missing, malformed or invalid "token_id"');
        }
        if (!isset($jsonData['auth_key']) || !\is_string($jsonData['auth_key'])) {
            throw new InvalidGrantException('missing, malformed or invalid "auth_key"');
        }
        if (!isset($jsonData['user_id']) || !\is_string($jsonData['user_id'])) {
            throw new InvalidGrantException('missing, malformed or invalid "user_id"');
        }
        if (!isset($jsonData['client_id']) || !\is_string($jsonData['client_id'])) {
            throw new InvalidGrantException('missing, malformed or invalid "client_id"');
        }
        if (!isset($jsonData['scope']) || !\is_string($jsonData['scope'])) {
            throw new InvalidGrantException('missing, malformed or invalid "scope"');
        }
        if (!isset($jsonData['expires_at']) || !\is_string($jsonData['expires_at'])) {
            throw new InvalidGrantException('missing, malformed or invalid "expires_at"');
        }

        return new self(
            $jsonData['token_id'],
            $jsonData['auth_key'],
            $jsonData['user_id'],
            $jsonData['client_id'],
            new Scope($jsonData['scope']),
            new DateTimeImmutable($jsonData['expires_at'])
        );
    }

    public function tokenId(): string
    {
        return $this->tokenId;
    }

    public function authKey(): string
    {
        return $this->authKey;
    }

    public function userId(): string
    {
        return $this->userId;
    }

    public function clientId(): string
    {
        return $this->clientId;
    }

    public function scope(): Scope
    {
        return $this->scope;
    }

    public function expiresAt(): DateTimeImmutable
    {
        return $this->expiresAt;
    }

    public function json(): string
    {
        return Json::encode(
            [
                'type' => self::TOKEN_TYPE,
                'token_id' => $this->tokenId,
                'auth_key' => $this->authKey,
                'user_id' => $this->userId,
                'client_id' => $this->clientId,
                'scope' => (string) $this->scope,
                'expires_at' => $this->expiresAt->format(DateTimeImmutable::ATOM),
            ]
        );
    }
}
